"""ModelhubCredential Implementation"""

# pylint: disable=line-too-long

import base64
import json
import time
from typing import Optional

import httpx

from autorag.utilities.logger import get_logger

logger = get_logger()


class ModelhubCredential:
    """
    Modelhub Class.

    This class instantiates a Modelhub client to perform various operations using our internal Modelhub's SDK.
    Example:
    .. code-block:: python

        from autorag.core.credential import ModelhubCredential

        modelhub_credential = ModelhubCredential(
            client_id="modelhub-client",
            client_secret="xxx",
            token="xxx"
        )

    """

    AUTH_URL = "https://auth.sprint.autonomize.dev/realms/autonomize/protocol/openid-connect/token"

    def __init__(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        token: Optional[str] = None,
    ):
        """
        Initialize the ModelhubCredential instance.

        Args:
            client_id (Optional[str]): The client ID for authentication.
            client_secret (Optional[str]): The client secret for authentication.
            token (Optional[str]): An existing JWT token for authentication.

        Raises:
            ValueError: If neither (client_id and client_secret) nor token is provided.
        """

        if client_id is None or client_secret is None:
            if token is None:
                raise ValueError(
                    "Either (`client_id` and `client_secret`) or `token` must be provided."
                )
            logger.warning(
                "It is recommended to provide `client_id` and `client_secret` over `token` "
                "because token gets expired within 24 hours. "
                "You can use `ModelhubCredential.is_token_expired(your_token)` to check if it has expired or not."
            )

        self._client_id = client_id
        self._client_secret = client_secret
        self._token = token

    @staticmethod
    def is_token_expired(token: str) -> bool:
        """
        Checks whether the provided JWT token is expired or not.

        Args:
            token (str): The JWT token to check if its expired or not.

        Returns:
            bool: True if token has expired, otherwise False.

        Raises:
            ValueError: When the token provided is not base64
        """

        # Split the token by `.` to get it in list
        b64_components = token.split(".")

        # The token should have three components: header, payload and signature.
        if len(b64_components) != 3:
            raise ValueError("Ill-formatted token provided. Please recheck your token.")

        # Divide it into, header, payload and signature,
        # but we don't need header and signature.
        _, payload_b64, _ = b64_components

        payload_json = base64.urlsafe_b64decode(payload_b64 + "==").decode("utf-8")
        payload = json.loads(payload_json)

        # Get expiry time in Unix-epoch time.
        token_expiry_time = payload["exp"]

        return token_expiry_time < int(time.time())

    def get_token(self) -> str:
        """
        Obtains a JWT token for authorization if token is not provided or has expired.

        Returns:
            str: JWT token.
        """
        if not self._token or ModelhubCredential.is_token_expired(self._token):
            if not self._client_id or not self._client_secret:
                raise ValueError(
                    "client_id and client_secret must be provided to fetch JWT token."
                )

            with httpx.Client() as client:
                response = client.post(
                    self.AUTH_URL,
                    headers={"Content-Type": "application/x-www-form-urlencoded"},
                    data={
                        "grant_type": "client_credentials",
                        "client_id": self._client_id,
                        "client_secret": self._client_secret,
                        "scope": "openid",
                    },
                )

                if response.status_code == 401:
                    raise ValueError(
                        "Client Error `401 Unauthorized`: client_id or client_secret is invalid."
                    )
                response.raise_for_status()
                self._token = response.json()["id_token"]

        # Ensure token is not None before returning
        if not self._token:
            raise ValueError("Token could not be retrieved.")  # pragma: no cover
        return self._token

    async def aget_token(self) -> str:
        """
        Asynchronously obtains a JWT token for authorization if token is not provided or has expired.

        Returns:
            str: JWT token.
        """
        if not self._token or ModelhubCredential.is_token_expired(self._token):
            if not self._client_id or not self._client_secret:
                raise ValueError(
                    "client_id and client_secret must be provided to fetch JWT token."
                )

            async with httpx.AsyncClient() as client:
                response = await client.post(
                    self.AUTH_URL,
                    headers={"Content-Type": "application/x-www-form-urlencoded"},
                    data={
                        "grant_type": "client_credentials",
                        "client_id": self._client_id,
                        "client_secret": self._client_secret,
                        "scope": "openid",
                    },
                )

                if response.status_code == 401:
                    raise ValueError(
                        "Client Error `401 Unauthorized`: client_id or client_secret is invalid."
                    )
                response.raise_for_status()
                self._token = response.json()["id_token"]

        # Ensure token is not None before returning
        if not self._token:
            raise ValueError("Token could not be retrieved.")  # pragma: no cover
        return self._token
